#include "jpeg_global.h"

/*
  cos ׻

  coeff[x][y] = cos( (2x+1)y / (2*MCU_SIZE) )
*/
double coeff[MCU_SIZE][MCU_SIZE];

/*
  doublecharѴ
*/
unsigned char dbl2char( double d )
{
  if ( d < 0 ) d = 0;
  if ( d > 255 ) d = 255;

  return( d );
}

/*
  cos()η׻

  coeff[x][y] = cos( (2x+1)y / (2*MCU_SIZE) )
*/
void calcCoeff()
{
  Coord p;

  for ( p.y = 0 ; p.y < MCU_SIZE ; ( p.y )++ )
    for ( p.x = 0 ; p.x < MCU_SIZE ; ( p.x )++ )
      coeff[p.x][p.y] = cos( ( ( 2 * p.x + 1 ) * p.y * M_PI ) / ( 2 * MCU_SIZE ) );
}

/*
  RGB->YUVѴ
*/
void rgb2yuv( DrawingArea* draw, double **yuv )
{
  Palet p;
  int gx, gy;
  double *y, *u, *v;
  guchar* pixel;

  y = yuv[Y_IDX];
  u = yuv[U_IDX];
  v = yuv[V_IDX];
  pixel = draw->pixbuf_p;

  for ( gy = 0 ; gy < draw->height ; gy++ ) {
    for ( gx = 0 ; gx < draw->width ; gx++ ) {
      s_point( &pixel, &p );
      *y++ =   ( 0.299 * (double)p.red )  + ( 0.587 * (double)p.green )  + ( 0.114 * (double)p.blue );
      *u++ = - ( 0.1687 * (double)p.red ) - ( 0.3313 * (double)p.green ) + ( 0.5 * (double)p.blue );
      *v++ =   ( 0.5 * (double)p.red )    - ( 0.4187 * (double)p.green ) - ( 0.0813 * (double)p.blue );
    }
    pixel += draw->pixbuf_xmod;
  }
}

/*
  YUV->RGBѴ
*/
void yuv2rgb( DrawingArea* draw, double **yuv )
{
  Palet p;
  int gx, gy;
  double *y, *u, *v;
  guchar* pixel;

  y = yuv[Y_IDX];
  u = yuv[U_IDX];
  v = yuv[V_IDX];
  pixel = draw->pixbuf_p;

  for ( gy = 0 ; gy < draw->height ; gy++ ) {
    for ( gx = 0 ; gx < draw->width ; gx++ ) {
      p.red   = dbl2char( *y                    + ( 1.4020 * *v )  );
      p.green = dbl2char( *y - ( 0.34414 * *u ) - ( 0.71414 * *v ) );
      p.blue  = dbl2char( *y + ( 1.772 * *u ) );
      s_pset( &pixel, &p );
      y++;
      u++;
      v++;
    }
    pixel += draw->pixbuf_xmod;
  }
  gdk_pixbuf_render_to_drawable( draw->pixbuf, draw->pixmap, draw->drawing_area->style->fg_gc[GTK_STATE_NORMAL],
				 0, 0, 0, 0, draw->width, draw->height, GDK_RGB_DITHER_NORMAL, 0, 0);
}

/*
  DCT׻(Xʬ¤˷׻)

                   n
  s_uy(u,y) = C(u) yuv(x,y)coeff(x,u) ( n=MCU_SIZE )
                  x=0

  double* yuv     : YUVǡ
  double s_uy[][] : s(u,y)
  Coord mcu_coord : MCU֥åθ߰
  Coord g_size    : 
  Coord skip      : ְ礭
*/
void dct_x( double* yuv, double s_uy[MCU_SIZE][MCU_SIZE], Coord* mcu_coord, Coord* g_size, Coord* skip )
{
  int u, x, y;
  double s_uv;         /* yuv(x,y)coeff(x,u) */
  double* current_yuv; /* yuv(x,y) */

  for ( y = 0 ; y < MCU_SIZE ; y++ ) {
    for ( u = 0 ; u < MCU_SIZE ; u++ ) {
      current_yuv = yuv + ( ( mcu_coord->y * MCU_SIZE + y ) * skip->y * g_size->x ) + ( mcu_coord->x * MCU_SIZE * skip->x );
      s_uv = 0;
      if ( ( mcu_coord->y * MCU_SIZE + y ) * skip->y < g_size->y ) {
	for ( x = 0 ; x < MCU_SIZE ; x++ ) {
	  if ( ( mcu_coord->x * MCU_SIZE + x ) * skip->x < g_size->x )
	    s_uv += *current_yuv * coeff[x][u];
	  current_yuv += skip->x;
	}
	if ( u == 0 ) s_uv /= sqrt( 2 );
      }
      s_uy[u][y] = s_uv;
    }
  }
}

/*
  DCT׻

                       n
  mcu(u,v) = (2/n)C(v) s_uy(u,y)coeff(y,v) ( n=MCU_SIZE )
                      y=0

  double s_uy[][] : s(u,y)
  Mcu* mcu        : MCU֥å
*/
void dct_y( double s_uy[MCU_SIZE][MCU_SIZE], Mcu* mcu )
{
  int y, u, v;
  double s_uv; /* s_uy(u,y)coeff(y,v) */

  for ( v = 0 ; v < MCU_SIZE ; v++ ) {
    for ( u = 0 ; u < MCU_SIZE ; u++ ) {
      s_uv = 0;
      for ( y = 0 ; y < MCU_SIZE ; y++ )
	s_uv += s_uy[u][y] * coeff[y][v];
      if ( v == 0 ) s_uv /= sqrt( 2 );
      ( mcu->d )[u][v] = s_uv * 2 / MCU_SIZE + ( ( s_uv >= 0 ) ? 0.5 : -0.5 );
    }
  }
}

/*
  YUV->MCUѴ(ְޤ)

  double* yuv  : YUVǡ
  Mcu* mcu     : DCTѴ̳Ǽ
  Coord g_size : 
  Coord skip   : ְ礭

         : mcuǸμΥݥ
*/
Mcu* mcu_encode( double* yuv, Mcu* mcu, Coord* g_size, Coord* skip )
{
  Coord mcu_size;                  /* MCUο(X,Yʬ) */
  Coord mcu_coord;                 /* MCU */
  double s_uy[MCU_SIZE][MCU_SIZE]; /* s_uy(u,y)ʬ */

  mcu_size.x = g_size->x / ( MCU_SIZE * skip->x );
  if ( g_size->x % ( MCU_SIZE * skip->x ) != 0 )
    ( mcu_size.x )++;
  mcu_size.y = g_size->y / ( MCU_SIZE * skip->y );
  if ( g_size->y % ( MCU_SIZE * skip->y ) != 0 )
    ( mcu_size.y )++;

  for ( mcu_coord.y = 0 ; mcu_coord.y < mcu_size.y ; ( mcu_coord.y )++ ) {
    for ( mcu_coord.x = 0 ; mcu_coord.x < mcu_size.x ; ( mcu_coord.x )++ ) {
      dct_x( yuv, s_uy, &mcu_coord, g_size, skip );
      dct_y( s_uy, mcu );
      mcu++;
    }
  }

  return( mcu );
}

/*
  YUVǡν񤭹

  double* yuv  : YUVǡ
  double data  : ϤYUVǡ
  int x,y      : 񤭹߰
  Coord g_size : 
  Coord skip   : ְ礭
*/
void fill_yuv( double* yuv, double data, int x, int y, Coord* g_size, Coord* skip )
{
  int i, j;

  for ( j = 0 ; j < skip->y ; j++ ) {
    if ( y + j >= g_size->y ) break;
    for ( i = 0 ; i < skip->x ; i++ ) {
      if ( x + i >= g_size->x ) break;
      yuv[ ( ( y + j ) * g_size->x ) + x + i ] = data;
    }
  }
}

/*
  IDCT׻

                   n
  yuv(x,y) = (2/n) C(u)s_uy(u,y)coeff(x,u) ( n=MCU_SIZE )
                  u=0

  double s_uy[][] : s(u,y)
  double* yuv     : YUVǡ
  Coord mcu_coord : MCU֥åθ߰
  Coord g_size    : 
  Coord skip      : ְ礭
*/
void idct_x( double s_uy[MCU_SIZE][MCU_SIZE], double *yuv, Coord* mcu_coord, Coord* g_size, Coord* skip )
{
  int u, x, y;
  int yuv_x, yuv_y; /*  */
  double s_uv;      /* C(u)s_uy(u,y)coeff(x,u) */

  for ( y = 0 ; y < MCU_SIZE ; y++ ) {
    if ( ( yuv_y = ( mcu_coord->y * MCU_SIZE * skip->y ) + ( y * skip->y ) ) >= g_size->y ) break;
    for ( x = 0 ; x < MCU_SIZE ; x++ ) {
      if ( ( yuv_x = ( mcu_coord->x * MCU_SIZE * skip->x ) + ( x * skip->x ) ) >= g_size->x ) break;
      s_uv = 0;
      for ( u = 0 ; u < MCU_SIZE ; u++ ) {
	double d = s_uy[u][y] * coeff[x][u];
	if ( u == 0 ) d /= sqrt( 2 );
	s_uv += d;
      }
      fill_yuv( yuv, s_uv * 2 / MCU_SIZE, yuv_x, yuv_y, g_size, skip );
    }
  }
}

/*
  IDCT׻(Vʬ¤˷׻)

               n
  s_uy(u,y) =  C(v)mcu(u,v)coeff(y,v) ( n=MCU_SIZE )
              v=0

  Mcu* mcu        : MCU֥å
  double s_uy[][] : s(u,y)
*/
void idct_y( Mcu* mcu, double s_uy[MCU_SIZE][MCU_SIZE] )
{
  int u, v, y;
  double s_uv; /* C(v)mcu(u,v)coeff(y,v) */

  for ( y = 0 ; y < MCU_SIZE ; y++ ) {
    for ( u = 0 ; u < MCU_SIZE ; u++ ) {
      s_uv = 0;
      for ( v = 0 ; v < MCU_SIZE ; v++ ) {
	double d = ( mcu->d )[u][v] * coeff[y][v];
	if ( v == 0 ) d /= sqrt( 2 );
	s_uv += d;
      }
      s_uy[u][y] = s_uv;
    }
  }
}

/*
  MCU->YUVѴ

  Mcu* mcu     : DCTѴ̳Ǽ
  double* yuv  : YUVǡ
  Coord g_size : 
  Coord skip   : ְ礭

*/
void mcu_decode( Mcu* mcu, double* yuv, Coord* g_size, Coord* skip )
{
  Coord mcu_size;                  /* MCUο(X,Yʬ) */
  Coord mcu_coord;                 /* MCU */
  double s_uy[MCU_SIZE][MCU_SIZE]; /* s_uy(u,y)ʬ */

  mcu_size.x = g_size->x / ( MCU_SIZE * skip->x );
  if ( g_size->x % ( MCU_SIZE * skip->x ) != 0 )
    ( mcu_size.x )++;
  mcu_size.y = g_size->y / ( MCU_SIZE * skip->y );
  if ( g_size->y % ( MCU_SIZE * skip->y ) != 0 )
    ( mcu_size.y )++;

  for ( mcu_coord.y = 0 ; mcu_coord.y < mcu_size.y ; ( mcu_coord.y )++ ) {
    for ( mcu_coord.x = 0 ; mcu_coord.x < mcu_size.x ; ( mcu_coord.x )++ ) {
      idct_y( mcu, s_uy );
      idct_x( s_uy, yuv, &mcu_coord, g_size, skip );
      mcu++;
    }
  }
}

/*
  եܥå
*/
void drawFillBox( DrawingArea* draw, int x, int y, Coord* size, Palet* palet )
{
  Coord p;

  for ( p.y = y ; p.y < y + size->y ; ( p.y )++ ) {
    if( p.y >= draw->height ) break;
    for ( p.x = x ; p.x < x + size->x ; ( p.x )++ ) {
      if( p.x >= draw->width ) break;
      pset( draw, p.x, p.y, palet );
    }
  }
}
/*
  MCUѴ̤
*/
void showMCU( DrawingArea* draw, Mcu* mcu, Coord* skip )
{
  Palet palet;
  Coord mcu_size;
  Coord mcu_coord;
  Coord p;

  mcu_size.x = draw->width / ( MCU_SIZE * skip->x );
  if ( draw->width % ( MCU_SIZE * skip->x ) != 0 )
    ( mcu_size.x )++;
  mcu_size.y = draw->height / ( MCU_SIZE * skip->y );
  if ( draw->height % ( MCU_SIZE * skip->y ) != 0 )
    ( mcu_size.y )++;

  for ( mcu_coord.y = 0 ; mcu_coord.y < mcu_size.y ; ( mcu_coord.y )++ ) {
    for ( mcu_coord.x = 0 ; mcu_coord.x < mcu_size.x ; ( mcu_coord.x )++ ) {
      for ( p.y = 0 ; p.y < MCU_SIZE ; ( p.y )++ ) {
	for ( p.x = 0 ; p.x < MCU_SIZE ; ( p.x )++ ) {
	  palet.red = palet.green = palet.blue = dbl2char( abs( ( mcu->d )[p.x][p.y] ) );
	  drawFillBox( draw, ( mcu_coord.x * MCU_SIZE + p.x ) * skip->x, ( mcu_coord.y * MCU_SIZE + p.y ) * skip->y, skip, &palet );
	}
      }
      mcu++;
    }
  }
  gdk_pixbuf_render_to_drawable( draw->pixbuf, draw->pixmap, draw->drawing_area->style->fg_gc[GTK_STATE_NORMAL],
				 0, 0, 0, 0, draw->width, draw->height, GDK_RGB_DITHER_NORMAL, 0, 0);
}
